@@ -1,6 +1,6 @@
 /*****************************************************************************
 MiniSat -- Copyright (c) 2003-2006, Niklas Een, Niklas Sorensson
-CryptoMiniSat -- Copyright (c) 2009 Mate Soos
+CryptoMiniSat -- Copyright (C) 2009-2020 Authors of CryptoMiniSat, see AUTHORS file
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -21,27 +21,42 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
 ******************************************************************************/
 
-#ifndef DIMACSPARSER_H
-#define DIMACSPARSER_H
+#pragma once
 
 #include <string.h>
 #include "streambuffer.h"
-#include "cryptominisat5/cryptominisat.h"
+#include "solvertypesmini.h"
 #include <cstdlib>
 #include <cmath>
+#include <sstream>
+#include <iostream>
+#include <iomanip>
+#include <vector>
+#include <fstream>
+#include <complex>
+#include <cassert>
 
-using namespace CMSat;
 using std::vector;
+using std::cout;
+using std::endl;
 
-template <class C>
+namespace CMSat {
+
+template <class C, class S>
 class DimacsParser
 {
     public:
-        DimacsParser(SATSolver* solver, const std::string* debugLib, unsigned _verbosity);
+        DimacsParser(S* solver, const std::string* debugLib, unsigned _verbosity);
 
-        template <class T> bool parse_DIMACS(T input_stream, const bool strict_header);
-        uint64_t max_var = std::numeric_limits<uint64_t>::max();
+        template <class T> bool parse_DIMACS(
+            T input_stpeam,
+            const bool strict_header,
+            uint32_t offset_vars = 0);
+        uint64_t max_var = numeric_limits<uint64_t>::max();
         vector<uint32_t> sampling_vars;
+        bool sampling_vars_found = false;
+        vector<double> weights;
+        uint32_t must_mult_exp2 = 0;
         const std::string dimacs_spec = "http://www.satcompetition.org/2009/format-benchmarks2009.html";
         const std::string please_read_dimacs = "\nPlease read DIMACS specification at http://www.satcompetition.org/2009/format-benchmarks2009.html";
 
@@ -50,17 +65,26 @@ class DimacsParser
         bool readClause(C& in);
         bool parse_and_add_clause(C& in);
         bool parse_and_add_xor_clause(C& in);
+        #ifdef ENABLE_BNN
+        bool parse_and_add_bnn_clause(C& in);
+        #endif
         bool match(C& in, const char* str);
-        bool printHeader(C& in);
+        bool parse_header(C& in);
         bool parseComments(C& in, const std::string& str);
         std::string stringify(uint32_t x) const;
+        bool check_var(const uint32_t var);
+        bool parseWeight(C& in);
+
+        #ifdef DEBUG_DIMACSPARSER_CMS
         bool parse_solve_simp_comment(C& in, const bool solve);
         void write_solution_to_debuglib_file(const lbool ret) const;
+        #endif
+
         bool parseIndependentSet(C& in);
         std::string get_debuglib_fname() const;
 
 
-        SATSolver* solver;
+        S* solver;
         std::string debugLib;
         unsigned verbosity;
 
@@ -75,6 +99,7 @@ class DimacsParser
         bool header_found = false;
         int num_header_vars = 0;
         int num_header_cls = 0;
+        uint32_t offset_vars = 0;
 
         //Reduce temp overhead
         vector<Lit> lits;
@@ -82,23 +107,15 @@ class DimacsParser
 
         size_t norm_clauses_added = 0;
         size_t xor_clauses_added = 0;
+        #ifdef ENABLE_BNN
+        size_t bnn_clauses_added = 0;
+        #endif
 };
 
-#include <sstream>
-#include <iostream>
-#include <iomanip>
-#include <vector>
-#include <fstream>
-#include <complex>
-#include <cassert>
 
-using std::vector;
-using std::cout;
-using std::endl;
-
-template<class C>
-DimacsParser<C>::DimacsParser(
-    SATSolver* _solver
+template<class C, class S>
+DimacsParser<C, S>::DimacsParser(
+    S* _solver
     , const std::string* _debugLib
     , unsigned _verbosity
 ):
@@ -111,16 +128,66 @@ DimacsParser<C>::DimacsParser(
     }
 }
 
-template<class C>
-std::string DimacsParser<C>::stringify(uint32_t x) const
+template<class C, class S>
+std::string DimacsParser<C, S>::stringify(uint32_t x) const
 {
     std::ostringstream o;
     o << x;
     return o.str();
 }
 
-template<class C>
-bool DimacsParser<C>::readClause(C& in)
+template<class C, class S>
+bool DimacsParser<C, S>::check_var(const uint32_t var)
+{
+    if (var > max_var) {
+        std::cerr
+        << "ERROR! "
+        << "Variable requested is too large for DIMACS parser parameter: "
+        << var << endl
+        << "--> At line " << lineNum+1
+        << please_read_dimacs
+        << endl;
+        return false;
+    }
+
+    if (var >= (1ULL<<28)) {
+        std::cerr
+        << "ERROR! "
+        << "Variable requested is far too large: " << var + 1 << endl
+        << "--> At line " << lineNum+1
+        << please_read_dimacs
+        << endl;
+        return false;
+    }
+
+    if (strict_header && !header_found) {
+        std::cerr
+        << "ERROR! "
+        << "DIMACS header ('p cnf vars cls') never found!" << endl;
+        return false;
+    }
+
+    if ((int)var >= num_header_vars && strict_header) {
+        std::cerr
+        << "ERROR! "
+        << "Variable requested is larger than the header told us." << endl
+        << " -> var is : " << var + 1 << endl
+        << " -> header told us maximum will be : " << num_header_vars << endl
+        << " -> At line " << lineNum+1
+        << endl;
+        return false;
+    }
+
+    if (var >= solver->nVars()) {
+        assert(!strict_header);
+        solver->new_vars(var - solver->nVars() +1);
+    }
+
+    return true;
+}
+
+template<class C, class S>
+bool DimacsParser<C, S>::readClause(C& in)
 {
     int32_t parsed_lit;
     uint32_t var;
@@ -133,51 +200,12 @@ bool DimacsParser<C>::readClause(C& in)
         }
 
         var = std::abs(parsed_lit)-1;
+        var += offset_vars;
 
-        if (var > max_var) {
-            std::cerr
-            << "ERROR! "
-            << "Variable requested is too large for DIMACS parser parameter: "
-            << var << endl
-            << "--> At line " << lineNum+1
-            << please_read_dimacs
-            << endl;
+        if (!check_var(var)) {
             return false;
         }
 
-        if (var >= (1ULL<<28)) {
-            std::cerr
-            << "ERROR! "
-            << "Variable requested is far too large: " << var + 1 << endl
-            << "--> At line " << lineNum+1
-            << please_read_dimacs
-            << endl;
-            return false;
-        }
-
-        if (strict_header && !header_found) {
-            std::cerr
-            << "ERROR! "
-            << "DIMACS header ('p cnf vars cls') never found!" << endl;
-            return false;
-        }
-
-        if ((int)var >= num_header_vars && strict_header) {
-            std::cerr
-            << "ERROR! "
-            << "Variable requested is larger than the header told us." << endl
-            << " -> var is : " << var + 1 << endl
-            << " -> header told us maximum will be : " << num_header_vars << endl
-            << " -> At line " << lineNum+1
-            << endl;
-            return false;
-        }
-
-        if (var >= solver->nVars()) {
-            assert(!strict_header);
-            solver->new_vars(var - solver->nVars() +1);
-        }
-
         lits.push_back( (parsed_lit > 0) ? Lit(var, false) : Lit(var, true) );
         if (*in != ' ') {
             std::cerr
@@ -194,8 +222,8 @@ bool DimacsParser<C>::readClause(C& in)
     return true;
 }
 
-template<class C>
-bool DimacsParser<C>::match(C& in, const char* str)
+template<class C, class S>
+bool DimacsParser<C, S>::match(C& in, const char* str)
 {
     for (; *str != 0; ++str, ++in)
         if (*str != *in)
@@ -203,10 +231,50 @@ bool DimacsParser<C>::match(C& in, const char* str)
     return true;
 }
 
-template<class C>
-bool DimacsParser<C>::printHeader(C& in)
+#ifdef DEBUG_DIMACSPARSER_CMS
+template<class C, class S>
+bool DimacsParser<C, S>::parseWeight(C& in)
 {
-    if (match(in, "p cnf")) {
+    if (match(in, "w ")) {
+        int32_t slit;
+        double weight;
+        if (in.parseInt(slit, lineNum)
+            && in.parseDouble(weight, lineNum)
+        ) {
+            if (slit == 0) {
+                cout << "ERROR: Cannot define weight of literal 0!" << endl;
+                exit(-1);
+            }
+            uint32_t var = std::abs(slit)-1;
+            bool sign = slit < 0;
+            Lit lit = Lit(var, sign);
+            solver->set_var_weight(lit, weight);
+            //cout << "lit: " << lit << " weight: " << std::setprecision(12) << weight << endl;
+            if (weight < 0) {
+                cout << "ERROR: while definint weight, variable " << var+1 << " has is negative weight: " << weight << " -- line " << lineNum << endl;
+                exit(-1);
+            }
+            return true;
+        } else {
+            cout << "ERROR: weight is incorrect on line " << lineNum << endl;
+            exit(-1);
+        }
+    } else {
+        cout << "ERROR: weight is not given on line " << lineNum << endl;
+        exit(-1);
+    }
+    return true;
+}
+#endif
+
+template<class C, class S>
+bool DimacsParser<C, S>::parse_header(C& in)
+{
+    ++in;
+    in.skipWhitespace();
+    std::string str;
+    in.parseString(str);
+    if (str == "cnf") {
         if (header_found && strict_header) {
             std::cerr << "ERROR: CNF header ('p cnf vars cls') found twice in file! Exiting." << endl;
             exit(-1);
@@ -230,6 +298,7 @@ bool DimacsParser<C>::printHeader(C& in)
             std::cerr << "ERROR: Number of clauses in header cannot be less than 0" << endl;
             return false;
         }
+        num_header_vars += offset_vars;
 
         if (solver->nVars() < (size_t)num_header_vars) {
             solver->new_vars(num_header_vars-solver->nVars());
@@ -244,7 +313,7 @@ bool DimacsParser<C>::printHeader(C& in)
         << std::dec
         << ")"
         << " At line " << lineNum+1
-        << "' in the header, at line " << lineNum+1
+        << "' in the header"
         << please_read_dimacs
         << endl;
         return false;
@@ -253,15 +322,16 @@ bool DimacsParser<C>::printHeader(C& in)
     return true;
 }
 
-template<class C>
-std::string DimacsParser<C>::get_debuglib_fname() const
+template<class C, class S>
+std::string DimacsParser<C, S>::get_debuglib_fname() const
 {
     std::string sol_fname = debugLib + "-debugLibPart" + stringify(debugLibPart) +".output";
     return sol_fname;
 }
 
-template<class C>
-bool DimacsParser<C>::parse_solve_simp_comment(C& in, const bool solve)
+#ifdef DEBUG_DIMACSPARSER_CMS
+template<class C, class S>
+bool DimacsParser<C, S>::parse_solve_simp_comment(C& in, const bool solve)
 {
     vector<Lit> assumps;
     in.skipWhitespace();
@@ -307,8 +377,8 @@ bool DimacsParser<C>::parse_solve_simp_comment(C& in, const bool solve)
     return true;
 }
 
-template<class C>
-void DimacsParser<C>::write_solution_to_debuglib_file(const lbool ret) const
+template<class C, class S>
+void DimacsParser<C, S>::write_solution_to_debuglib_file(const lbool ret) const
 {
     //Open file for writing
     std::string s = get_debuglib_fname();
@@ -345,10 +415,12 @@ void DimacsParser<C>::write_solution_to_debuglib_file(const lbool ret) const
     }
     partFile.close();
 }
+#endif
 
-template<class C>
-bool DimacsParser<C>::parseComments(C& in, const std::string& str)
+template<class C, class S>
+bool DimacsParser<C, S>::parseComments(C& in, const std::string& str)
 {
+    #ifdef DEBUG_DIMACSPARSER_CMS
     if (!debugLib.empty() && str.substr(0, 13) == "Solver::solve") {
         if (!parse_solve_simp_comment(in, true)) {
             return false;
@@ -357,7 +429,37 @@ bool DimacsParser<C>::parseComments(C& in, const std::string& str)
         if (!parse_solve_simp_comment(in, false)) {
             return false;
         }
-    } else if (!debugLib.empty() && str == "Solver::new_var()") {
+    } else
+    #endif
+    if (str == "red") {
+        in.skipWhitespace();
+        lits.clear();
+        if (!readClause(in)) return false;
+        solver->add_red_clause(lits);
+    }
+    if (str == "MUST") {
+        std::string str2;
+        in.skipWhitespace();
+        in.parseString(str2);
+        if (str2 != "MULTIPLY") {
+            cout << "ERROR: expected 'MULTIPLY' after 'MUST'" << endl;
+            return false;
+        }
+        in.skipWhitespace();
+        in.parseString(str2);
+        if (str2 != "BY") {
+            cout << "ERROR: expected 'BY' after 'MUST MULTIPLY'" << endl;
+            return false;
+        }
+        in.skipWhitespace();
+        assert(*in == '2');++in;
+        assert(*in == '*');++in;
+        assert(*in == '*');++in;
+        if (!in.parseInt(must_mult_exp2, lineNum)) {
+            return false;
+        }
+    }
+    if (!debugLib.empty() && str == "Solver::new_var()") {
         solver->new_var();
 
         if (verbosity >= 6) {
@@ -375,9 +477,22 @@ bool DimacsParser<C>::parseComments(C& in, const std::string& str)
             cout << "c Parsed Solver::new_vars( " << n << " )" << endl;
         }
     } else if (str == "ind") {
+        sampling_vars_found = true;
         if (!parseIndependentSet(in)) {
             return false;
         }
+    } else if (str == "p") {
+        in.skipWhitespace();
+        std::string str2;
+        in.parseString(str2);
+        if (str2 == "show") {
+            in.skipWhitespace();
+            sampling_vars_found = true;
+            if (!parseIndependentSet(in)) { return false; }
+        } else {
+            cout << "ERROR, 'c p' followed by unknown text: '" << str2 << "'" << endl;
+            exit(-1);
+        }
     } else {
         if (verbosity >= 6) {
             cout
@@ -391,8 +506,8 @@ bool DimacsParser<C>::parseComments(C& in, const std::string& str)
     return true;
 }
 
-template<class C>
-bool DimacsParser<C>::parse_and_add_clause(C& in)
+template<class C, class S>
+bool DimacsParser<C, S>::parse_and_add_clause(C& in)
 {
     lits.clear();
     if (!readClause(in)) {
@@ -408,8 +523,67 @@ bool DimacsParser<C>::parse_and_add_clause(C& in)
     return true;
 }
 
-template<class C>
-bool DimacsParser<C>::parse_and_add_xor_clause(C& in)
+#ifdef ENABLE_BNN
+// This parses a threshold constraint of the form:
+// b lit1..litn 0 cutoff 0 [output_lit]
+// where output_lit is optional and if missing is assumed to be TRUE
+// basically, lit1+lit2+litN >= cutoff  <=> output_lit=TRUE
+// where lit1 is 1 if it's TRUE and 0 otherwise
+template<class C, class S>
+bool DimacsParser<C, S>::parse_and_add_bnn_clause(C& in)
+{
+    // Read in inputs to BNN
+    lits.clear();
+    if (!readClause(in)) {
+        return false;
+    }
+    if (lits.empty()) {
+        std::cerr
+        << "ERROR! "
+        << "BNN constraint has empty set of inputs" << endl
+        << "--> At line " << lineNum+1
+        << endl;
+        return false;
+    }
+
+    // Read cutoff
+    int32_t cutoff;
+    if (!in.parseInt(cutoff, lineNum)) {
+        return false;
+    }
+
+    in.skipWhitespace();
+
+    Lit out = lit_Undef;
+    if (*in != '\n') {
+        // Read in output var
+        int32_t parsed_lit;
+        if (!in.parseInt(parsed_lit, lineNum)) {
+            return false;
+        }
+        assert(parsed_lit != 0);
+        uint32_t var = std::abs(parsed_lit)-1;
+        var += offset_vars;
+
+        //off-by-one internally.
+        if (!check_var(var)) {
+            return false;
+        }
+        out = Lit(var, parsed_lit < 0);
+    }
+
+    // Line finished
+    in.skipLine();
+    lineNum++;
+
+    solver->add_bnn_clause(lits, cutoff, out);
+    bnn_clauses_added++;
+    return true;
+}
+#endif
+
+template<class C, class S>
+bool DimacsParser<C, S>::parse_and_add_xor_clause(C& in)
 {
     lits.clear();
     if (!readClause(in)) {
@@ -435,8 +609,8 @@ bool DimacsParser<C>::parse_and_add_xor_clause(C& in)
     return true;
 }
 
-template<class C>
-bool DimacsParser<C>::parse_DIMACS_main(C& in)
+template<class C, class S>
+bool DimacsParser<C, S>::parse_DIMACS_main(C& in)
 {
     std::string str;
 
@@ -446,14 +620,23 @@ bool DimacsParser<C>::parse_DIMACS_main(C& in)
         case EOF:
             return true;
         case 'p':
-            if (!printHeader(in)) {
+            if (!parse_header(in)) {
                 return false;
             }
             in.skipLine();
             lineNum++;
             break;
-        case 'c':
+
+        #ifdef DEBUG_DIMACSPARSER_CMS
         case 'w':
+            if (!parseWeight(in)) {
+                return false;
+            }
+            in.skipLine();
+            lineNum++;
+            break;
+        #endif
+        case 'c':
             ++in;
             in.parseString(str);
             if (!parseComments(in, str)) {
@@ -466,12 +649,25 @@ bool DimacsParser<C>::parse_DIMACS_main(C& in)
                 return false;
             }
             break;
+        case 'b':
+            #ifdef ENABLE_BNN
+            ++in;
+            if (!parse_and_add_bnn_clause(in)) {
+                return false;
+            }
+            #else
+            std::cout << "ERROR: BNN encounered but not enabled in parsing. Exiting." << endl;
+            exit(-1);
+            #endif
+            break;
         case '\n':
-            std::cerr
-            << "c WARNING: Empty line at line number " << lineNum+1
-            << " -- this is not part of the DIMACS specifications ("
-            << dimacs_spec << "). Ignoring."
-            << endl;
+            if (verbosity) {
+                std::cout
+                << "c WARNING: Empty line at line number " << lineNum+1
+                << " -- this is not part of the DIMACS specifications ("
+                << dimacs_spec << "). Ignoring."
+                << endl;
+            }
             in.skipLine();
             lineNum++;
             break;
@@ -486,12 +682,16 @@ bool DimacsParser<C>::parse_DIMACS_main(C& in)
     return true;
 }
 
-template <class C>
+template <class C, class S>
 template <class T>
-bool DimacsParser<C>::parse_DIMACS(T input_stream, const bool _strict_header)
+bool DimacsParser<C, S>::parse_DIMACS(
+    T input_stream,
+    const bool _strict_header,
+    uint32_t _offset_vars)
 {
     debugLibPart = 1;
     strict_header = _strict_header;
+    offset_vars = _offset_vars;
     const uint32_t origNumVars = solver->nVars();
 
     C in(input_stream);
@@ -503,6 +703,9 @@ bool DimacsParser<C>::parse_DIMACS(T input_stream, const bool _strict_header)
         cout
         << "c -- clauses added: " << norm_clauses_added << endl
         << "c -- xor clauses added: " << xor_clauses_added << endl
+        #ifdef ENABLE_BNN
+        << "c -- bnn clauses added: " << bnn_clauses_added << endl
+        #endif
         << "c -- vars added " << (solver->nVars() - origNumVars)
         << endl;
     }
@@ -510,8 +713,8 @@ bool DimacsParser<C>::parse_DIMACS(T input_stream, const bool _strict_header)
     return true;
 }
 
-template <class C>
-bool DimacsParser<C>::parseIndependentSet(C& in)
+template <class C, class S>
+bool DimacsParser<C, S>::parseIndependentSet(C& in)
 {
     int32_t parsed_lit;
     for (;;) {
@@ -527,4 +730,4 @@ bool DimacsParser<C>::parseIndependentSet(C& in)
     return true;
 }
 
-#endif //DIMACSPARSER_H
+}
